# coding=utf-8

from django.contrib.auth.models import User
from django.db.models import Q

from engines.models import Cluster, StepOperation, Scale
from engines.models import COMPONENT_OPERATION
from data.models import Dataset, DataInstance, DataInstanceOperation, DataEngine
from data.models import INSTANCE_STATUS as DATA_INSTANCE_STATUS
from clouds.models import OPERATION_STATUS, INSTANCE_STATUS, INSTANCE_OPERATION
from clouds.models import Instance, InstanceOperation

def _cluster_info(obj):
    return {
        'id': obj.id,
        'uuid' : obj.uuid,
        'name' : obj.name,
        'scale' : obj.scale,
        'engines' : obj.engines,
        'steps' : obj.steps,
        'remedy_script_todo' : obj.remedy_script_todo,
        'remark' : obj.remark,
        'public' : obj.public,
        'owner' : obj.owner,
        'created_time' : obj.created_time,
        'built_time' : obj.built_time,
        'status' : obj.status,
        'status_name': INSTANCE_STATUS(obj.status).name,        
        'deleting' : obj.deleting,
        'portal' : obj.portal if obj.ready else None,
    }

def get_cluster_list(req_user: User) -> list:
    objs = Cluster.objects.filter(deleting=False).order_by("-id")   
    if not req_user.is_superuser:
        objs = objs.filter(Q(owner=req_user) | Q(public=True))
    return [_cluster_info(obj) for obj in objs]

def _get_cluster_obj(req_user: User, c_id) -> Cluster:
    objs = Cluster.objects.filter(id=c_id, deleting=False)
    if not req_user.is_superuser:
        objs = objs.filter(Q(owner=req_user) | Q(public=True)) 
    if objs.exists():
        return objs.first()
    return None

def get_cluster_info(req_user: User, c_id) -> dict:
    obj = _get_cluster_obj(req_user, c_id)
    if obj != None:
        return _cluster_info(obj)
    return None

def operate_cluster(req_user: User, c_id: int, operation: str) -> bool:
    ops1 = ("start","reboot","shutdown","poweroff","remedy")
    ops2 = ("scale_in", "scale_out")
    ops3 = ("destroy", )
    if (operation not in ops1) and (operation not in ops2) and (operation not in ops3):
        return False

    objs = Cluster.objects.filter(id=c_id, deleting=False)
    if not req_user.is_superuser:
        objs = objs.filter(Q(owner=req_user) | Q(public=True)) 
    if objs.exists():
        obj = objs.first()
        if operation in ops1:
            op = INSTANCE_OPERATION(operation)
            obj.get_operation_model()(
                    target=obj,
                    operation=op.value,
                    status=OPERATION_STATUS.running.value
                ).save()
        elif operation == "scale_out" :  
            obj.scale_one_step()
        elif operation == "scale_in" : 
            obj.steps.last().delete()
        elif operation == "destroy" :
            obj.delete()
        return True

    return False


def add_cluster(owner_id, name, scale, engines, public=False, remedy_script_todo=None, remark=None):
    obj = Cluster()
    obj.owner_id = owner_id
    obj.name = name
    obj.scale_id = scale
    obj.public = public
    obj.remedy_script_todo = remedy_script_todo
    obj.remark = remark
    obj.save()
    obj.engines.set(engines)


def _get_cluster_instance_info(obj, require_vnc=False):
    vnc_url = None
    try:
        if require_vnc:
            vnc_url = obj.vnc_url
    except Exception as e:
        pass

    return {
            "id": obj.id,
            "ipv4": obj.ipv4,
            "ipv6": obj.ipv6,
            "cloud_id": obj.cloud_id,
            "cloud_name": obj.cloud.name,
            "uuid": obj.uuid,
            "image_id": obj.image.id,
            "image_name": obj.image.name,
            "template_id": obj.template.id,
            "template_name": obj.template.name,
            "hostname": obj.hostname,
            "remedy_script_todo": obj.remedy_script_todo,
            "created_time": obj.created_time,
            "built_time": obj.built_time,
            "remark": obj.remark,
            "owner": obj.owner,
            "status": obj.status,
            "status_name": INSTANCE_STATUS(obj.status).name,            
            "deleting": obj.deleting,
            "vnc_url": vnc_url,
            
    }

def get_cluster_instances(req_user: User, c_id, require_vnc=False) -> list:
    obj = _get_cluster_obj(req_user, c_id)
    if obj != None:
        return [_get_cluster_instance_info(instance, require_vnc=require_vnc) for instance in obj.get_instances()   ]     
    return []

def get_cluster_instance_info(req_user:User, instance_id, require_vnc=False) -> dict:
    objs = Instance.objects.filter(id=instance_id)
    if objs.exists():
        return _get_cluster_instance_info(objs.first(), require_vnc=require_vnc)
    return None


def operate_cluster_instance(req_user: User, c_id: int, instance_id: int, operation: str) -> bool:
    ops1 = ("toggle","delete")
    if operation not in ops1 :
        return False

    objs = Instance.objects.filter(id=instance_id)
    if not req_user.is_superuser:
        objs = objs.filter(owner=req_user) 
    if objs.exists():
        obj = objs.first()
        if operation in ops1:
            InstanceOperation(
                target=obj,
                operation=INSTANCE_OPERATION.poweroff.value if obj.status==INSTANCE_STATUS.active.value else INSTANCE_OPERATION.start.value
            ).save()       
        elif operation == "delete" :  
            obj.delete()
        return True
    return False



def _scale_info(obj):
    return {
        "id": obj.id,
        "name": obj.name,
        # "engines": [{"id": e.id, "name":e.name} for e in obj.available_engines]
    }


def get_scale_list():
    objs = Scale.objects.filter(enabled=True)
    return [_scale_info(obj) for obj in objs]
    

def get_available_engines(scale_id):
    scales = Scale.objects.filter(id=scale_id)
    required = ["HDFS","HIVE","MAPREDUCE2","YARN","TEZ","HBASE","PIG",
                "ZOOKEEPER","AMBARI_METRICS","SPARK","SPARK2","AMBARI_INFRA",
                "ASTROSERV","GSTORE","EVENTDB","PIFLOW","SIMBA","REDIS","SLIDER","SPARKJS",
                "ZEPPELIN","HUE"]
    if scales.exists():
        obj = scales.first()
        res = []
        if obj.stack:
            for e in obj.available_engines:
                if e.name in required:
                    res.append({"id": e.id, "name":e.name, "required":True})
                else:
                    res.append({"id": e.id, "name":e.name, "required":False})
        return res
    return []


def _dataset_info(obj):
    return {
            'id': obj.id,
            'name': obj.name,
            'uuid': obj.uuid,
            'source': obj.source,
            'uri': obj.uri,
            'type': obj.type,
            'type_name': obj.type_name,
            'size': obj.size,
            'description': obj.description,
            'public': obj.public,
            'enabled': obj.enabled,
            'modified_time': obj.modified_time,
            'created_time': obj.created_time,
            'remark': obj.remark,
            'owner': obj.owner,
        }


def get_dataset_list(req_user: User) -> list:
    objs = Dataset.objects.filter(enabled=True).order_by("-id") 
    if not req_user.is_superuser:
        objs = objs.filter(Q(owner=req_user) | Q(public=True))
    return [_dataset_info(obj) for obj in objs]


def get_dataset_info(dt_id):
    objs = Dataset.objects.filter(id=dt_id, enabled=True)
    if objs.exists():
        return _dataset_info(objs.first())
    return None


def _engine_info(obj):
    return {
        'id': obj.id,
        'uuid': obj.uuid,
        'name': obj.name,
        'description': obj.description,
    }


def _data_instance_info(obj):
    query_url = ""
    if obj.cluster.ready:
        query_url = obj.cluster.portal.replace("8080","8888").rstrip("/")
        if obj.engine:
            engine_name = obj.engine.name.strip().lower()
            if engine_name == "hive":
                query_url += "/beeswax/#query"
            elif engine_name == "hdfs":
                query_url += "/filebrowser/#" + obj.uri
            elif engine_name == "hbase":
                query_url += "/hbase/"
            elif obj.uri.strip():
                query_url = obj.uri.strip()
                if query_url.startswith("http"):
                    if query_url.count("+") > 0:
                        query_url = query_url.split("+")[0].strip()
                else:
                    query_url = ""
    return {
        'id': obj.id,
        'uuid': obj.uuid,
        'name': obj.name,
        'remedy_script_todo': obj.remedy_script_todo,
        'created_time': obj.created_time,
        'built_time': obj.built_time,
        'owner': obj.owner,
        'status': obj.status,
        'status_name': DATA_INSTANCE_STATUS(obj.status).name,
        'remark': obj.remark,
        'deleting': obj.deleting,
        'dataset': _dataset_info(obj.dataset),      
        'cluster': _cluster_info(obj.cluster),
        'engine': _engine_info(obj.engine),
        'engine_lower_name':obj.engine.name.strip().lower() if obj.engine else None,
        'uri': obj.uri,
        'query_url': query_url,
    }



def get_data_instance_list(req_user: User, cluster_id) -> list:
    objs = DataInstance.objects.filter(cluster_id = cluster_id).order_by("-id") 
    if not req_user.is_superuser:
        objs = objs.filter(owner=req_user)
    return [_data_instance_info(obj) for obj in objs]

def get_data_instance_info(di_id):
    objs = DataInstance.objects.filter(id=di_id)
    if objs.exists():
        return _data_instance_info(objs.first())
    return None

def add_data_instance(owner_id, cluster_id, name, dataset_id, data_engine_id, remedy_script_todo=None, remark=None):
    obj = DataInstance()
    obj.owner_id = owner_id
    obj.name = name
    obj.dataset_id = dataset_id
    obj.cluster_id = cluster_id
    obj.engine_id = data_engine_id
    obj.remedy_script_todo = remedy_script_todo
    obj.remark = remark
    obj.save()

def delete_data_instance(req_user: User, di_id):
    objs = DataInstance.objects.filter(id=di_id, deleting=False)
    if not req_user.is_superuser:
        objs = objs.filter(owner=req_user)
    objs.delete()

# scispace operations
def _step_operation_info(obj):
    return {
    'id': obj.id,
    'name': str(obj),
    'operation': obj.operation,
    'operation_name': INSTANCE_OPERATION(obj.operation).name,
    'batch_uuid': obj.batch_uuid,
    'script': obj.script,
    'log': obj.log,
    'serial': obj.serial,
    'created_time': obj.created_time,
    'started_time': obj.started_time,
    'completed_time': obj.completed_time,
    'status': obj.status,
    'tidied': obj.tidied,
    'manual': obj.manual,
    'cluster': obj.target,
    }


def get_step_operation_list(cluster_id):
    cluster_obj = Cluster.objects.filter(id=cluster_id, deleting=False).first()
    objs = StepOperation.objects.filter(target__cluster = cluster_obj).exclude(status=OPERATION_STATUS.success.value)
    return [_step_operation_info(obj) for obj in objs]

def get_step_operation_info(cluster_id, op_id):
    cluster_obj = Cluster.objects.filter(id=cluster_id, deleting=False).first()
    objs = StepOperation.objects.filter(target__cluster = cluster_obj, id=op_id)
    if objs.exists():
        return _step_operation_info(objs.first())
    return None


# data instance operations
def _cloud_instance_operation_info(obj):
    return {
    'id': obj.id,
    'name': str(obj),
    'operation': obj.operation,
    'operation_name': INSTANCE_OPERATION(obj.operation).name,
    'batch_uuid': obj.batch_uuid,
    'script': obj.script,
    'serial': obj.serial,
    'created_time': obj.created_time,
    'started_time': obj.started_time,
    'completed_time': obj.completed_time,
    'status': obj.status,
    'tidied': obj.tidied,
    'manual': obj.manual,   
    'target': obj.target,
    'log': obj.log,
    }


def get_data_instance_operation_list(cluster_id):
    objs = InstanceOperation.objects.filter(script__contains=("mkdir -p /data/packone/"))
    return [_cloud_instance_operation_info(obj) for obj in objs]

def get_data_instance_operation_info(cluster_id,op_id):
    objs = InstanceOperation.objects.filter(id=op_id)
    if objs.exists():
        return _cloud_instance_operation_info(objs.first())
    return None

def _data_engine_info(obj):
    return {
        "id": obj.id,
        "engine": obj.engine,
        "engine_name": obj.engine.name,
        "uuid": obj.uuid,
        "uri_prefix": obj.uri_prefix,
        "description": obj.description
    }

def get_cluster_data_engine_list(cluster_id, dataset_id):    
    if not cluster_id or not dataset_id:
        return []
    qs = DataEngine.objects.filter(
            engine__in = Cluster.objects.get(pk=cluster_id).engines.all(),
            type=Dataset.objects.get(pk=dataset_id).type
        ).order_by('-pk')
    return [_data_engine_info(obj) for obj in qs]
        


from datetime import timedelta, datetime
from django.db.models import Sum, Count
import data.models as models
def get_data_metrics(cluster_id):
    today = datetime.now()
    begin_day = today - timedelta(days=365)
    state = {
        # 'cluster':{},
        'dataset':{},
        'instance':{},
        # 'space':{}
    }
    space = Cluster.objects.filter(id=cluster_id).first()
    state['dataset']['type_cnt']=[{'type': models.DATASET_TYPE(pair['type']).name, 'cnt': pair['cnt']} for pair in list(models.Dataset.objects.values('type').annotate(cnt=Count('id')))]
    state['dataset']['type_size']=[{'type': models.DATASET_TYPE(pair['type']).name, 'size': pair['size']} for pair in list(models.Dataset.objects.all().values('type').annotate(size=Sum('size')))]
    state['dataset']['public_size']=list(models.Dataset.objects.values('public').annotate(size=Sum('size')))
    state['dataset']['owner_size']=list(models.Dataset.objects.values('owner__username').annotate(size=Sum('size')).order_by('-size')[0:10])
    state['dataset']['month_size']=list(models.Dataset.objects.filter(modified_time__range=(begin_day, today)).values('modified_time__year', 'modified_time__month').annotate(size=Sum('size')).order_by('modified_time__year', 'modified_time__month'))
    state['instance']['status_cnt']=[{'status': models.INSTANCE_STATUS(pair['status']).name, 'cnt': pair['cnt']} for pair in list(models.DataInstance.objects.filter(cluster=space).values('status').annotate(cnt=Count('id')))]
    state['instance']['type_cnt']=[{'dataset__type': models.DATASET_TYPE(pair['dataset__type']).name, 'cnt': pair['cnt']} for pair in list(models.DataInstance.objects.filter(cluster=space).values('dataset__type').annotate(cnt=Count('id')).order_by('-cnt')[0:10])]
    state['instance']['dataset_cnt']=list(models.DataInstance.objects.filter(cluster=space).values('dataset__name').annotate(cnt=Count('id')).order_by('-cnt')[0:10])
    
    return state

def get_hosts_metrics(req_user: User, c_id) -> dict:
    obj = _get_cluster_obj(req_user, c_id)
    metrics = obj.metrics
    return metrics



from threading import Thread
import time 
from django.core.management.base import BaseCommand
from clouds.models import Instance

class Command(BaseCommand):
    help = 'monitor cloud resource periodly'
    def handle(self, *args, **options):
        while True:
            for instance in Instance.objects.exclude(uuid=None):
                Thread(target=instance.monitor).start()
                time.sleep(0.1)
            time.sleep(300)



from threading import Thread
import datetime, time 
from django.core.management.base import BaseCommand, CommandError
from data.models import Space
class Command(BaseCommand):
    help = 'monitor data instances periodly'
    def handle(self, *args, **options):
        while True:
            for s in Space.objects.all():
                p=s.pilot
                for e in p.engines.all():
                    pes=e.status(p)
                    dis=s.datainstance_set.filter(engine__component__in=e.components.all())
                    for di in dis:
                        di.status=pes
                        di.save()
            time.sleep(300)


import paramiko
from scp import SCPClient
from django.conf import settings
import json
from io import StringIO
import calendar
import time
from django.urls import reverse
from django.utils.html import format_html

def get_url(obj):
    label= obj.name if getattr(obj,'name',False) else str(obj)
    url=reverse('admin:{}_{}_change'.format(obj._meta.app_label,obj._meta.model_name),args=(obj.pk,))
    return '<a href='+url+'>'+label+'</a>'

def get_formated_url(obj):
    return format_html(get_url(obj))

def get_current_timestamp():
    ts = calendar.timegm(time.gmtime())
    return ts

def get_refer_GET_parameter(request,name=None):
    from urllib import parse
    params=parse.parse_qs(parse.urlparse(request.META['HTTP_REFERER']).query)
    if not name:
        return params
    if name in params:
        return params[name][0]
    return None

def remedy_script_tidy(script, supervisor_operations):
    stats=[s.rstrip() for s in script.split('\n')]
    begin_i=0
    current_i=0
    length=len(stats)
    operations=[]
    while current_i < length:
        if not stats[current_i] or stats[current_i] in supervisor_operations:
            if current_i>begin_i:
                operations.append('\n'.join(stats[begin_i:current_i]))
            if stats[current_i]:
                operations.append(stats[current_i])  
            begin_i=current_i+1
        elif current_i==length-1:
            operations.append('\n'.join(stats[begin_i:current_i+1]))
        current_i+=1
    return operations

def remedy_script_hosts_add(hosts,overwrite=False):
    if overwrite: return "echo '{hosts}' >/etc/hosts".format(hosts=hosts)
    return "echo '{hosts}' >>/etc/hosts".format(hosts=hosts)

#TODO only support linux
def remedy_script_hosts_remove(hosts):
    return ';'.join(["sed -i '/{}/d' /etc/hosts".format(h) for h in hosts.split('\n') if h])
#TODO only support linux
def remedy_script_hosts_remove_from(tag):
    return "sed -i '/{}/,$d' /etc/hosts".format(tag)
#TODO only support linux
def remedy_script_hostname(hostname):
    return "hostnamectl set-hostname {hostname}".format(hostname=hostname)

#TODO make filesystem type chooseable
def remedy_script_mount_add(mount):
    return "mkfs.xfs {mount.dev}>/dev/null 2>&1\n" \
    "mv {mount.point} {mount.point}.packone.old >/dev/null 2>&1\n" \
    "mkdir -p {mount.point}\n" \
    "mount {mount.dev} {mount.point}\n" \
    "mv {mount.point}.packone.old/* {mount.point}/ >/dev/null 2>&1\n" \
    "rm -rf {mount.point}.packone.old\n" \
    "echo '{mount.dev} {mount.point} xfs defaults 0 2'>>/etc/fstab". \
    format(mount=mount)

def remedy_script_mount_remove(mount):
    return "sed -i '/{}/d' /etc/fstab".format('{mount.dev} {mount.point}'.format(mount=mount).replace('/', '\/'))

def remedy_image_ambari_agent(nameserver='8.8.8.8', repo_url='https://public-repo-1.hortonworks.com/ambari/centos7/2.x/updates/2.7.3.0/ambari.repo'):
    return "echo 'nameserver {}'>>/etc/resolv.conf\ncurl -Ssl {} -o /etc/yum.repos.d/ambari.repo\n\nyum -q -y install ambari-agent 2>&1".format(nameserver,repo_url)

def remedy_image_ambari_server():
    return 'yum -q -y install ambari-server 2>&1\n\n' \
        'ambari-server setup -s >/dev/null\n\n' \
        'ambari-server start'

from paramiko import RSAKey
def gen_ssh_key():
    key=RSAKey.generate(2048)
    private_key_file=StringIO()
    key.write_private_key(private_key_file)
    pri=private_key_file.getvalue()
    private_key_file.close()
    pub=key.get_base64()
    pub='ssh-rsa '+pub
    return pub, pri

def get_pub_key(private_key):
    private_key_file=StringIO(private_key)
    key=paramiko.RSAKey.from_private_key(private_key_file)
    private_key_file.close()
    pub=key.get_base64()
    pub='ssh-rsa '+pub
    return pub

class SSH:
    def __init__(self,host,username='root',password=None,private_key=None,port=None):
        self.username=username
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if private_key:
            private_key_file=StringIO(private_key)
            private_key=paramiko.RSAKey.from_private_key(private_key_file)
            private_key_file.close()
        mustend = time.time() + 900
        e=Exception()
        while time.time() < mustend:
            time.sleep(5)
            try:
                self.client.connect(
                    host,
                    username=username,
                    password=password,
                    pkey=private_key,
                    timeout=None
                )
            # except paramiko.AuthenticationException as ex:
            #     raise ex
            except paramiko.ssh_exception.NoValidConnectionsError as ex:
                continue
            except Exception as ex:
                # import traceback
                # traceback.print_tb(ex.__traceback__)
                print(ex)
                e=ex
                continue
            else:
                return
        raise e

    def exec_batch(self, cmd):
        ftp = self.client.open_sftp()
        file=ftp.file('/tmp/packone.bash', "w", -1)
        file.write(cmd)
        file.flush()
        ftp.close()
        cmd='sudo -uroot ' if self.username!='root' else ''
        cmd+='bash /tmp/packone.bash'
        stdin, stdout, stderr = self.client.exec_command(cmd)
        err='\n'.join(stderr.readlines())
        if err: err='STDERR:\n'+err
        out='STDOUT:\n'+'\n'.join(stdout.readlines())
        if out.find('ERROR: Exiting with exit code 1')!=-1:
            err+=out
            out=''
        return out, err

    def close(self):
        self.client.close()


from keystoneauth1 import loading, session
from novaclient import client as nova_client
from cinderclient import client as cinder_client
from glanceclient import client as glance_client
from novaclient.exceptions import NotFound as NovaNotFound
from cinderclient.exceptions import NotFound as CinderNotFound
from uuid import uuid4
import time

class Driver(object):
    def __init__(self, cloud):
        self._cloud=cloud
        self._credential=cloud.platform_credential
        auth=loading.get_plugin_loader('password').load_from_options(
            auth_url=self._credential['auth_url'],
            username=self._credential['username'],
            password=self._credential['password'],
            project_name=self._credential['project_name']
        )
        kwargs={
            'version': self._credential['api_version'],
            'session': session.Session(auth=auth)
        }
        self._nova_client=nova_client.Client(**kwargs)
        self._cinder_client=cinder_client.Client(**kwargs)
        self._glance_client=glance_client.Client(**kwargs)
        self.instances=InstanceManager(self)
        self.volumes=VolumeManager(self)
        self.flavors=self._nova_client.flavors
        self.keypairs=self._nova_client.keypairs#create, delete
        self.images=self._glance_client.images

class InstanceManager(object):
    mountable_status=['ACTIVE','SHUTDOWN']
    def __init__(self, driver):
        self.driver=driver
        self._manager=driver._nova_client.servers
        self.get=self._manager.get
        self.list=self._manager.list
    def create(self, image_id, template_id, remark='', **kwargs):
        ins=self._manager.create(
            name=remark,
            image=image_id,
            flavor=template_id,
            security_groups=[self.driver._credential['security_group']],
            nics=[{'net-id':self.driver._credential['net-id']}],
            key_name=self.driver._cloud._key_name,
            userdata="#cloud-config\n" \
            "debug: True\n" \
            "ssh_pwauth: True\n" \
            "disable_root: false\n" \
            "runcmd:\n" \
            "- sed -i'.orig' -e's/without-password/yes/' /etc/ssh/sshd_config\n" \
            "- service sshd restart"
        )
        mustend = time.time() + 600
        while time.time() < mustend:
            ins=self.get(ins.id)
            if 'provider' in ins.addresses:
                break
            time.sleep(5)
        return ins
    def delete(self, id):
        try:
            return self._manager.delete(id)
        except NovaNotFound as e:
            print(e)
    def force_delete(self, id):
        return self.delete(id)
    def get_status(self, id):
        ins=self.get(id)
        return ins.status
#TODO ins.stop() return 202
class VolumeManager(object):
    def __init__(self, driver):
        self.driver=driver
        self._manager=driver._cinder_client.volumes
        self.get=self._manager.get
        self.list=self._manager.list
    def create(self, size, remark=''):
        volume=self._manager.create(
            name=remark,
            size=size
        )
        mustend = time.time() + 60
        while time.time() < mustend:
            volume=self.get(volume.id)
            if volume.status == 'available': break
            time.sleep(5)
        return volume
    def delete(self, volume_id):
        mustend = time.time() + 60
        while time.time() < mustend:
            try:
                volume=self._manager.get(volume_id)
            except CinderNotFound as e:
                print(e)
                return
            if volume.status == 'available': break
            time.sleep(5)
        return volume.delete()
    def mount(self, volume_id, instance_id):
        self.driver._nova_client.volumes.create_server_volume(server_id=instance_id, volume_id=volume_id)
        mustend = time.time() + 60
        while time.time() < mustend:
            volume=self.get(volume_id)
            if volume.status == 'in-use': break
            time.sleep(5)
        return volume
    def unmount(self, volume_id, instance_id):
        mustend = time.time() + 60
        while time.time() < mustend:
            ins=self.driver.instances.get(instance_id)
            if ins.status == 'SHUTOFF': break
            time.sleep(5)
        return self.driver._nova_client.volumes.delete_server_volume(server_id=instance_id, volume_id=volume_id)


from django.contrib import admin
from . import models
from clouds.base.admin import StaticModelAdmin, OwnershipModelAdmin, OperationAdmin, OperatableAdminMixin, AutoModelAdmin, M2MOperationAdmin
from clouds.models import Cloud, Group, InstanceBlueprint, InstanceOperation, GroupOperation
from user.models import Balance
from django.utils.html import format_html
from django.urls import reverse
from django.db.models import Q
from django import forms
from dal import autocomplete
from user.utils import get_current_user, get_space
from clouds.utils import get_url, get_formated_url

@admin.register(models.Stack)
class StackAdmin(StaticModelAdmin):
    pass

@admin.register(models.Engine)
class EngineAdmin(StaticModelAdmin):
    pass

@admin.register(models.Scale)
class ScaleAdmin(StaticModelAdmin):
    list_filter = ('auto',)+StaticModelAdmin.list_filter
    def get_queryset_Q(self, request):
        return Q(pk__in=request.user.scales())

@admin.register(models.Cluster)
class ClusterAdmin(OwnershipModelAdmin,OperatableAdminMixin):
    def access(self, obj):
        if not obj.ready: return None
        return format_html('<a href="{}" target="_blank" class="button">Manage</a>'.format(obj.portal))
    def instances(self,object):
        return format_html('<br/>'.join([get_url(ins) for ins in object.get_instances()]))
    search_fields = ('name','scale__name')+OwnershipModelAdmin.search_fields
    list_filter = (('scale', admin.RelatedOnlyFieldListFilter),)+OwnershipModelAdmin.list_filter
    extra=('access','instances')
    def get_list_display_exclude(self, request, obj=None):
        if request.user.is_superuser: return ('instances',)
        return ('deleting','instances')
    def start(modeladmin, request, queryset):
        for cluster in queryset:
            models.ClusterOperation(
                target=cluster,
                operation=models.INSTANCE_OPERATION.start.value,
                status=models.OPERATION_STATUS.running.value
            ).save()
    start.short_description = "Start selected clusters"
    def stop(modeladmin, request, queryset):
        for cluster in queryset:
            models.ClusterOperation(
                target=cluster,
                operation=models.INSTANCE_OPERATION.poweroff.value,
                status=models.OPERATION_STATUS.running.value
            ).save()
    stop.short_description = "Stop selected clusters"
    def materialize(modeladmin, request, queryset):
        for cluster in queryset:
            if not '.bootstrap.' in cluster.scale.name: continue
            blueprints=[]
            for ins in cluster.get_instances():
                image, created=ins.get_or_create_image()
                blueprint, created=ins.cloud.instanceblueprint_set.get_or_create(
                    name=image.name,
                    cloud=ins.cloud,
                    template=ins.template,
                    image=image,
                    defaults={
                        'owner': ins.owner,
                        'volume_capacity': 500,
                        'remark':'auto created'
                    }
                )
                blueprints.append(blueprint)
            from .utils import remedy_scale_ambari_fast_init, remedy_scale_ambari_fast_scale_out, remedy_scale_ambari_fast_scale_in
            s, created=models.Scale.objects.get_or_create(
                name=cluster.name.replace('bootstrap','packone'),
                _remedy_script=remedy_scale_ambari_fast_init(),
                _remedy_script_scale_out=remedy_scale_ambari_fast_scale_out(),
                _remedy_script_scale_in=remedy_scale_ambari_fast_scale_in(),
                stack=cluster.scale.stack,
                public=True,
                owner=cluster.owner
            )
            s.init_blueprints.add(*blueprints)
            s.step_blueprints.add(*list(filter(lambda x: 'slave' in x.name, blueprints)))
    materialize.short_description = "Materialize the cluster as a scale"
    def scale_out(modeladmin, request, queryset):
        for cluster in queryset:
            cluster.scale_one_step()
    scale_out.short_description = "Scale out one step"
    def scale_in(modeladmin, request, queryset):
        for cluster in queryset:
            cluster.steps.last().delete()
    scale_in.short_description = "Scale in one step"
    def destroy(modeladmin, request, queryset):
        for cluster in queryset:
            cluster.delete()
    destroy.short_description = "Destroy selected clusters"
    def import_engine(modeladmin, request, queryset):
        for cluster in queryset:
            cluster.import_engine()
    import_engine.short_description = "Refresh engines from selected clusters"
    actions=[start,stop,materialize,scale_out,scale_in,destroy,import_engine]
    def has_delete_permission(self, request, obj=None):
        return False
    def get_form_field_queryset_Q(self, db_field, request):
        if db_field.name == 'scale': return Q(pk__in=request.user.scales())
        return None
    def get_queryset_Q(self, request):
        return Q(pk__in=request.user.clusters())
    
@admin.register(models.ClusterOperation)
class ClusterOperationAdmin(M2MOperationAdmin):
    def has_module_permission(self, request):
        return False

@admin.register(models.StepOperation)
class StepOperationAdmin(M2MOperationAdmin):
    def _target(self, obj):
        return  format_html(get_url(obj.cluster)+'/'+get_url(obj.target))
    def get_queryset_Q(self, request):
        return Q(target__cluster=get_space()) & ~Q(status=models.OPERATION_STATUS.success.value)
    def has_add_permission(self, request, obj=None):
        return False
    def has_change_permission(self, request, obj=None):
        return False
    def has_view_permission(self, request, obj=None):
        return True
    def has_delete_permission(self, request, obj=None):
        return True
    def has_module_permission(self, request):
        return True


from django.urls import re_path,include
from rest_framework import routers
from . import views

ROUTER = routers.DefaultRouter()
ROUTER.register(r'datasources', views.DataSourceViewSet,basename='datasource')
ROUTER.register(r'datasets', views.DatasetViewSet,basename='dataset')
ROUTER.register(r'dataengines', views.DataEngineViewSet,basename='dataengine')
ROUTER.register(r'datainstances', views.DataInstanceViewSet,basename='datainstance')
ROUTER.register(r'datainstanceoperations', views.DataInstanceOperationViewSet,basename='datainstanceoperation')

urlpatterns = [
    re_path(r'^', include(ROUTER.urls)),
    re_path(
        r'^datainstanceengine-autocomplete/$',
        views.DataInstanceEngineAutocompleteView.as_view(),
        name='datainstanceengine-autocomplete',
    ),
]



from django.contrib import admin
from django.db.models import Q
from . import models
from clouds.admin import AutoModelAdmin, OwnershipModelAdmin
from clouds.models import Cloud

@admin.register(models.Profile)
class ProfileAdmin(OwnershipModelAdmin):
    search_fields = ('organization',)
    def has_delete_permission(self, request, obj=None):
        if not obj: return False
        if request.user.is_superuser: return True
        if obj.enabled: return False
        return True

class EnabledProfileGuardedAdmin(AutoModelAdmin):
    def enabled(self,obj):
        return obj.profile.enabled
    enabled.boolean = True
    extra=('enabled',)
    search_fields = ('profile__organization', 'profile__owner__username')
    def get_queryset_Q(self, request):
        return Q(profile__owner=request.user)

@admin.register(models.Balance)
class BalanceAdmin(EnabledProfileGuardedAdmin):#TODO add balanceZeroException in signal
    search_fields=('cloud__name', 'cloud___driver')+EnabledProfileGuardedAdmin.search_fields
    def has_change_permission(self, request, obj=None):
        return not obj or request.user==obj.cloud.owner
    def has_add_permission(self, request, obj=None):
        return Cloud.objects.filter(owner=request.user).exists()
    def has_delete_permission(self, request, obj=None):
        return not obj or obj.cloud.owner==request.user
    def get_queryset_Q(self, request):
        return super().get_queryset_Q(request)|Q(cloud__owner=request.user)
    def get_form_field_queryset_Q(self, db_field, request):
        if db_field.name=='cloud': return Q(owner=request.user)
        if db_field.name=='profile': return Q(enabled=True) #TODO use text instead of select to choose profile

@admin.register(models.Credential)
class CredentialAdmin(EnabledProfileGuardedAdmin):
    search_fields = ('ssh_user',)+EnabledProfileGuardedAdmin.search_fields
    def has_delete_permission(self, request, obj=None):
        if request.user.is_superuser: return True
        return False
    def has_add_permission(self, request, obj=None):
        return False
   



from django.dispatch import receiver
from django.db.models.signals import pre_save,post_save,pre_delete,post_delete,m2m_changed
from .import models
import datetime
from django.db import transaction
from multiprocessing.pool import ThreadPool
from threading import Thread
from clouds import utils
from clouds.signals import materialized, executed, monitored, destroyed, tidied, selected
from clouds.signals import tidy_operation, select_operation, executed
from clouds.models import InstanceOperation
from engines.models import Cluster

loading_instance_operations={}

@receiver(post_save, sender=models.DataInstance)
def materialize_data_instance(sender,instance,**kwargs):
    if not kwargs['created'] or instance.ready: return
    tmpdir="/data/packone/"+instance.uri_suffix
    scripts="mkdir -p "+tmpdir+' && cd '+tmpdir+'\n'
    scripts+=instance.dataset.remedy_script.format(dataset=instance.dataset, instance=instance)+'\n\n'
    io=instance.entry_host.remedy(scripts)
    loading_instance_operations[io.pk]=instance.pk

@receiver(executed, sender=InstanceOperation)
@transaction.atomic
def update_data_instance_status(sender,instance,**kwargs):
    if instance.pk not in loading_instance_operations: return
    for di in models.DataInstance.objects.filter(pk=loading_instance_operations[instance.pk]).select_for_update():
        del loading_instance_operations[instance.pk]
        di.built_time=datetime.datetime.now()
        di.status=models.INSTANCE_STATUS.running.value
        di.save()

@receiver(materialized, sender=models.DataInstance)
@receiver(tidied, sender=models.DataInstanceOperation)
@receiver(executed, sender=models.DataInstanceOperation)
def monitor_data_instance(sender, instance, **kwargs):
    if sender==models.DataInstanceOperation:
        if instance.status==models.OPERATION_STATUS.waiting.value: return
    #     instance=instance.target
    # if not instance.ready: return
    # Thread(target=instance.monitor).start()

post_save.connect(tidy_operation,sender=models.DataInstanceOperation)
monitored.connect(select_operation,sender=models.DataInstance)
@receiver(selected, sender=models.DataInstanceOperation)
def execute_operation(sender,instance,**kwargs):
    instance.execute()


"""packone URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf import settings
from django.contrib import admin
from django.urls import include, path, re_path
from django.views.static import serve
from rest_framework_simplejwt.views import TokenObtainPairView, TokenRefreshView

from data import views

admin.site.site_header = 'PackOne Console' # default: "Django Administration"
admin.site.index_title = 'Dashboard' # default: "Site administration"
admin.site.site_title = 'PackOne Console' # default: "Django site admin"
admin.site.site_url = None

urlpatterns = [
    # re_path(r'^', admin.site.urls),
    # re_path(r'^space/[0-9]+/', admin.site.urls),
    re_path(r'^jet/', include('jet.urls', 'jet')),  # Django JET URLS
    re_path(r'^jet/dashboard/', include('jet.dashboard.urls', 'jet-dashboard')),  # Django JET dashboard URLS
    path('api/user/', include('user.urls')),
    path('api/clouds/', include('clouds.urls')),
    path('api/engines/', include('engines.urls')),
    path('api/data/', include('data.urls')),

    re_path(r'^space/[0-9]+/api/data/state/', views.space_state),
    re_path(r'^static/(?P<path>.*)$', serve, {'document_root':settings.STATIC_ROOT}),
    re_path(r'^api/token/$', TokenObtainPairView.as_view(), name='token_obtain_pair'),
    re_path(r'^api/token/refresh/$', TokenRefreshView.as_view(), name='token_refresh'),
]


urlpatterns += [
    re_path(r'^', include('scispace.urls')),
    re_path(r'^admin/', admin.site.urls),
    # path('scispace/', include('scispace.urls')),

]


from django.db import models
from django.contrib.auth.models import User
from django.core.validators import MinValueValidator
from clouds.models import Cloud

#TODO allow user's teammates to access his/her resources
class Profile(models.Model):
    owner=models.ForeignKey(User,on_delete=models.CASCADE,editable=False)
    avatar=models.CharField(max_length=100,blank=True,null=True)
    organization=models.CharField(max_length=50, default='default')
    remark = models.CharField(blank=True,null=True,max_length=100)
    enabled=models.BooleanField(default=False)
    class Meta:
        unique_together = ('owner', 'organization')
    def __str__(self):
        return "{}/{}".format(self.owner,self.organization)

class Balance(models.Model): #todo add balance allocate/retrieve
    cloud=models.ForeignKey(Cloud,on_delete=models.CASCADE)
    profile=models.ForeignKey(Profile,on_delete=models.CASCADE)
    balance=models.IntegerField(validators=[MinValueValidator(0)],default=0)
    modified_time=models.DateTimeField(auto_now=True)
    class Meta:
        unique_together = ('profile', 'cloud')
    def __str__(self):
        return "{}/{}".format(self.profile, self.cloud)
        
def balances_of_user(self):
    return Balance.objects.filter(
        profile__owner=self,
        profile__enabled=True,
        balance__gt=0
    )
User.balances=balances_of_user

class Credential(models.Model):
    profile=models.OneToOneField(Profile,on_delete=models.CASCADE,editable=False)
    ssh_user=models.CharField(max_length=50,default='root',editable=False)
    ssh_public_key=models.CharField(max_length=5048,blank=True,null=True) #todo add length restrction
    ssh_passwd=models.CharField(max_length=50,blank=True,null=True)
    class Meta:
        unique_together = ('profile', 'ssh_user')
    def __str__(self):
        return "{}/{}".format(self.profile, self.ssh_user)


from rest_framework import viewsets
from rest_framework.permissions import IsAuthenticated, SAFE_METHODS
from django.db.models import Q, IntegerField, Case, When, Count
from clouds import permissions
from clouds.models import Image
from .models import Scale
from . import models
from . import serializers
from dal import autocomplete

class EngineViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.EngineSerializer
    filter_fields=('public',)
    permission_classes = (IsAuthenticated, permissions.IsOwnerOrAdminOrPublicReadOnly,)
    def get_queryset(self):
        queryset = models.Engine.objects.filter(Q(public=True) | Q(owner=self.request.user)).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class ScaleViewSet(viewsets.ModelViewSet):
    filter_fields=('public',)
    permission_classes = (IsAuthenticated,permissions.IsOwnerOrAdminOrPublicReadOnly,)
    http_method_names = ['get','post','delete','head','options']
    serializer_class = serializers.ScaleSerializer
    def get_queryset(self):
        queryset = models.Scale.objects.filter(Q(public=True) | Q(owner=self.request.user)).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)
    
class ClusterViewSet(viewsets.ModelViewSet):
    permission_classes = (IsAuthenticated,permissions.IsOwner,)
    serializer_class = serializers.ClusterSerializer
    def get_queryset(self):
        queryset = models.Cluster.objects.filter(owner=self.request.user).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class ClusterOperationViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.ClusterOperationSerializer
    http_method_names = ['get','post','head','options']
    def get_queryset(self):
        queryset = models.ClusterOperation.objects.filter(target__owner=self.request.user).order_by('-pk')
        return queryset


from rest_framework import serializers
from django.db.models import Q
from clouds.serializers import ImagePKField, InstanceBlueprintPKField
from . import models

class EngineSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    class Meta:
        model = models.Engine
        fields = '__all__'

class EnginePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.Engine.objects.filter(Q(public=True) | Q(owner=user))
        return queryset

class ScaleSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    init_blueprints = InstanceBlueprintPKField(many=True)
    step_blueprints = InstanceBlueprintPKField(many=True)
    instance_quantity = serializers.ReadOnlyField()
    cpu_quantity = serializers.ReadOnlyField()
    ram_quantity = serializers.ReadOnlyField()
    volume_capacity = serializers.ReadOnlyField()
    class Meta:
        model = models.Scale
        fields = '__all__'

class ScalePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.Scale.objects.filter(Q(public=True) | Q(owner=user))
        return queryset

class ClusterPKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        qs = models.Cluster.objects.all()
        if not user.is_superuser:
            qs = qs.filter(owner=user)
        return qs

class ClusterSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source='owner.username')
    scale=ScalePKField()
    engines=EnginePKField(many=True)
    portal = serializers.ReadOnlyField()
    class Meta:
        model = models.Cluster
        fields = '__all__'

class ClusterOperationSerializer(serializers.ModelSerializer):
    target=ClusterPKField()
    class Meta:
        model = models.ClusterOperation
        fields = '__all__' 


from ambari.client import Client

class Driver(object):
    def __init__(self, portal):
        self.portal=portal
        self.client=Client(portal)
    @property
    def stack_version(self):
        stack=self.client.cluster.stack
        return stack.name+'-'+stack.version
    def get_engine_host(self, engine):
        return self.client.cluster.hosts[-1].name
    @property
    def stack_engines(self):
        return self.client.stack.services
    @property
    def cluster_engines(self):
        return self.client.cluster.services
    def list_components(self, engine):
        return self.client.stack.get_service(engine).components
    @property
    def metrics(self):
        m=[]
        for h in self.client.cluster.hosts:
            m.append({h.name: h.metrics})
        return m


from . import models
from . import serializers
from rest_framework import viewsets
from django_filters import rest_framework as filters
from django.db.models import Q
from clouds import permissions
import clouds
from rest_framework.permissions import IsAuthenticated
from django.db import transaction
from rest_framework.permissions import SAFE_METHODS
from rest_framework.decorators import api_view
from django.db.models import Sum, Count
from rest_framework.response import Response
from datetime import timedelta, datetime
from dal import autocomplete
from engines.models import Engine,Cluster
from user.utils import get_space

@api_view(['GET'])
def space_state(request):
    today = datetime.now()
    begin_day = today - timedelta(days=365)
    state = {
        'cluster':{},
        'dataset':{},
        'instance':{},
        # 'space':{}
    }
    space=get_space()
    state['cluster']['metics']=space.metrics
    # state['dataset']['total_type']=set([d.type_name for d in models.Dataset.objects.all()])
    state['dataset']['type_cnt']=[{'type': models.DATASET_TYPE(pair['type']).name, 'cnt': pair['cnt']} for pair in list(models.Dataset.objects.values('type').annotate(cnt=Count('id')))]
    state['dataset']['type_size']=[{'type': models.DATASET_TYPE(pair['type']).name, 'size': pair['size']} for pair in list(models.Dataset.objects.all().values('type').annotate(size=Sum('size')))]
    # state['dataset']['total_size']=models.Dataset.objects.all().aggregate(total_size=Sum('size'))['total_size']
    state['dataset']['public_size']=list(models.Dataset.objects.values('public').annotate(size=Sum('size')))
    state['dataset']['owner_size']=list(models.Dataset.objects.values('owner__username').annotate(size=Sum('size')).order_by('-size')[0:10])
    state['dataset']['month_size']=list(models.Dataset.objects.filter(modified_time__range=(begin_day, today)).values('modified_time__year', 'modified_time__month').annotate(size=Sum('size')).order_by('modified_time__year', 'modified_time__month'))
    state['instance']['status_cnt']=[{'status': models.INSTANCE_STATUS(pair['status']).name, 'cnt': pair['cnt']} for pair in list(models.DataInstance.objects.filter(cluster=space).values('status').annotate(cnt=Count('id')))]
    state['instance']['type_cnt']=[{'dataset__type': models.DATASET_TYPE(pair['dataset__type']).name, 'cnt': pair['cnt']} for pair in list(models.DataInstance.objects.filter(cluster=space).values('dataset__type').annotate(cnt=Count('id')).order_by('-cnt')[0:10])]
    state['instance']['dataset_cnt']=list(models.DataInstance.objects.filter(cluster=space).values('dataset__name').annotate(cnt=Count('id')).order_by('-cnt')[0:10])
    # state['instance']['owner_cnt']=list(models.DataInstance.objects.values('owner__username').annotate(cnt=Count('id')).order_by('-cnt')[0:10])
    # state['instance']['total_cnt']=models.DataInstance.objects.all().count()
    
    return Response(state)

class DataInstanceEngineAutocompleteView(autocomplete.Select2QuerySetView):
    def get_queryset(self):
        if not self.request.user.is_authenticated:
            return models.DataEngine.objects.none()
        cluster_id = self.forwarded.get('cluster', None)
        dataset_id = self.forwarded.get('dataset', None)
        if not cluster_id or not dataset_id:
            return models.DataEngine.objects.none()
        qs = models.DataEngine.objects.filter(
            engine__in = Cluster.objects.get(pk=cluster_id).engines.all(),
            type=models.Dataset.objects.get(pk=dataset_id).type
        ).order_by('-pk')
        if self.q:
            qs = qs.filter(name__istartswith=self.q)
        return qs

#TODO finer permission control
class DataSourceViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.DataSourceSerializer
    filter_fields=('public',)
    ordering_fields = '__all__'
    search_fields = ('name','uri','remark','description')
    permission_classes = (IsAuthenticated,permissions.IsOwnerOrAdminOrPublicReadOnly,)
    # http_method_names = ['get','post','delete','head','options']
    def get_queryset(self):
        queryset = models.DataSource.objects.filter(Q(public=True) | Q(owner=self.request.user)).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class DatasetViewSet(viewsets.ModelViewSet):
    filter_fields=('public','source',)
    ordering_fields = '__all__'
    search_fields = ('name','remark','description')
    permission_classes = (IsAuthenticated,permissions.IsOwnerOrAdminOrPublicReadOnly,)
    #http_method_names = ['get','post','delete','head','options']
    def get_serializer_class(self):
        if self.request.method in SAFE_METHODS:
            return serializers.DatasetReadSerializer
        else:
            return serializers.DatasetSerializer
    def get_queryset(self):
        queryset = models.Dataset.objects.filter(Q(public=True) | Q(owner=self.request.user)).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class DataEngineViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.DataEngineSerializer
    filter_fields=('public',)
    permission_classes = (IsAuthenticated,permissions.IsOwnerOrAdminOrPublicReadOnly,)
    http_method_names = ['get','post','delete','head','options']
    def get_queryset(self):
        queryset = models.DataEngine.objects.filter(Q(public=True) | Q(owner=self.request.user)).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class DataInstanceViewSet(viewsets.ModelViewSet):
    filter_fields=('space','dataset','engine')
    ordering_fields = '__all__'
    search_fields = ('name','uri_suffix','remark')
    serializer_class = serializers.DataInstanceSerializer
    permission_classes = (IsAuthenticated,permissions.IsOwner,)
    http_method_names = ['get','post','delete','head','options']
    def get_queryset(self):
        queryset = models.DataInstance.objects.filter(owner=self.request.user).order_by('-pk')
        return queryset
    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

class DataInstanceOperationViewSet(viewsets.ModelViewSet):
    serializer_class = serializers.DataInstanceOperationSerializer
    http_method_names = ['get','post','head','options']
    def get_queryset(self):
        queryset = models.DataInstanceOperation.objects.filter(target__owner=self.request.user).order_by('-pk')
        return queryset


from setuptools import setup, find_packages

setup(
    name = "pk1",
    version = "0.3.3",
    keywords = ("pip", "packone"),
    description = "PackOne: Pack clouds, engines and data services into one light stack.",
    long_description = open('README.rst').read(),
    license = "Apache-2.0 Licence",
    url = "https://github.com/cas-bigdatalab/packone",
    author = "Excel Wang",
    author_email = "wanghj@cnic.com",
    packages = find_packages(exclude=['*.migrations']),
    include_package_data = True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires = ['httplib2==0.18.1','django-jet-opt==1.0.8','django==2.2.28','uwsgi==2.0.21','ambari==0.1.7','django-admin-auto==0.1.1','djangorestframework==3.10.3','djangorestframework-simplejwt==4.3.0','django-simpleui==4.0.3', 'django-filter==2.4.0', 'psycopg2-binary==2.8.6', 'coreapi==2.3.3', 'paramiko==2.7.2', 'scp==0.14.4', 'django-cors-headers==2.4.0', 'python-novaclient==17.7.0', 'python-cinderclient==7.4.1', 'python-glanceclient==3.6.0', 'django-autocomplete-light==3.8.2'],
    entry_points = {
        'console_scripts': [
            'pk1 = pk1.server:main',
        ]
    }
)



def remedy_scale_ambari_bootstrap(vdf_url=None):
    script="sed -i 's/hostname=localhost/hostname=master1.packone/g' /etc/ambari-agent/conf/ambari-agent.ini\n\n" \
        "ambari-agent start 2>&1\n\n" \
        'if [ `hostname` == "master1.packone" ]; then\n' \
        'yum -q -y install epel-release 2>&1\n' \
        'yum -q -y install python-pip 2>&1\n' \
        'pip -qq install ambari\n' \
        "fi\n\n" \
        'if [ `hostname` == "master1.packone" ]; then\n' \
        'yum -q -y install nmap-ncat 2>&1\n' \
        'while ! echo exit | nc localhost 8080; do sleep 3; done 2>&1\n' \
        'ambari localhost:8080 cluster create'
    if vdf_url: script+='_from_vdf '+vdf_url
    script+=' packone typical_triple master1.packone master2.packone slave.packone\nfi'
    return script

def remedy_scale_ambari_fast_init():
    return 'rm -rf /hadoop\n' \
        'mkdir -p /data/hadoop\n' \
        'ln -sf /data/hadoop /hadoop\n' \
        '#env\n' \
        "echo 'JAVA_HOME=/usr/jdk64/default'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'JRE_HOME=/usr/jdk64/default/jre'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'CLASS_PATH=.:$JAVA_HOME/lib:$JRE_HOME/lib'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'PATH=$JAVA_HOME/bin:$JRE_HOME/bin:$PATH'>>/etc/profile.d/packone-java.sh\n" \
        'reboot\n' \
        'if [ `hostname` == "master1.packone" ]; then\n' \
        '    while ! echo exit | nc localhost 8080; do sleep 3; done 2>&1\n' \
        '    pip -qq install -U ambari\n' \
        '    ambari master1:8080 service start\n' \
        'fi'

def remedy_scale_ambari_fast_scale_out():
    return "rm -rf /hadoop\n" \
        "mkdir -p /data/hadoop\n" \
        "ln -sf /data/hadoop /hadoop\n" \
        "#env\n" \
        "echo 'JAVA_HOME=/usr/jdk64/default'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'JRE_HOME=/usr/jdk64/default/jre'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'CLASS_PATH=.:$JAVA_HOME/lib:$JRE_HOME/lib'>>/etc/profile.d/packone-java.sh\n" \
        "echo 'PATH=$JAVA_HOME/bin:$JRE_HOME/bin:$PATH'>>/etc/profile.d/packone-java.sh\n" \
        "reboot\n" \
        "pip -qq install -U ambari\n" \
        'yum -q -y install nmap-ncat 2>&1\n' \
        'while ! echo exit | nc master1 8080; do sleep 3; done 2>&1\n' \
        "ambari master1:8080 host clone slave.packone `hostname`\n" \
        'ambari master1:8080 service start'

def remedy_scale_ambari_fast_scale_in():
    return 'if [ `hostname` == "master1.packone" ]; then\nambari master1:8080 host delete {hostname}\nfi'


"""
Django settings for packone project.

Generated by 'django-admin startproject' using Django 2.1.1.

For more information on this file, see
https://docs.djangoproject.com/en/2.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.1/ref/settings/
"""

import os
# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'vqvkx(j7fe#h#3(gm#t3e0mwjuy8_6skgu+qefgb!+7k7lx_su'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ['*']

# Application definition

INSTALLED_APPS = [
    'jet.dashboard',
    'jet',
    'dal',
    'dal_select2',
    'clouds.apps.CloudsConfig',
    'engines.apps.EnginesConfig',
    'data.apps.DataConfig',
    'user.apps.UserConfig',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'django_filters',
    'corsheaders',
    'scispace',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'user.utils.CurrentUserMiddleware'
]

ROOT_URLCONF = 'conf.urls'

JET_SIDE_MENU_COMPACT = True

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': ['templates',],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'conf.wsgi.application'

# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': 'packone_new',
        'USER': 'rabbit',
        'PASSWORD': 'rabbit',
        'HOST': 'rabbit-db-host',
        'PORT': '5432'
    },
    # 'default': {
    #     'ENGINE': 'django.db.backends.sqlite3',
    #     'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    #     'OPTIONS': {
    #         'timeout': 3600000,  # in seconds
    #     }
    # }
}

# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

USE_TZ = False
TIME_ZONE = 'Asia/Shanghai'

USE_I18N = True

USE_L10N = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/

STATIC_URL = '/static/'
# STATIC_ROOT = './static/'
STATICFILES_DIRS = [
    os.path.join(BASE_DIR, "static"),
]

PACKONE_LABEL='__packone__'
PACKONE_STACK_VDF_URLS=[
    {
        'name': 'HDP-2.6.4.0',
        'vdf': 'http://public-repo-1.hortonworks.com/HDP/centos7/2.x/updates/2.6.4.0/HDP-2.6.4.0-91.xml'
    },
    {
        'name': 'HDP-2.6.5.0',
        'vdf': 'http://public-repo-1.hortonworks.com/HDP/centos7/2.x/updates/2.6.5.0/HDP-2.6.5.0-292.xml'
    },
    {
        'name': 'HDP-3.1.0.0',
        'vdf': 'http://public-repo-1.hortonworks.com/HDP/centos7/3.x/updates/3.1.0.0/HDP-3.1.0.0-78.xml'
    }
]

REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    # or allow read-only access for unauthenticated users.
    # 'DEFAULT_METADATA_CLASS': 'clouds.drf_metadata.ReEnableRelatedFieldMetadata',
    'DEFAULT_PERMISSION_CLASSES': [
        # 'rest_framework.permissions.DjangoModelPermissionsOrAnonReadOnly',
        'rest_framework.permissions.IsAuthenticated'
    ],
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework_simplejwt.authentication.JWTAuthentication',
    ),
    'DEFAULT_PAGINATION_CLASS': 'clouds.pagination.PageSizeNumberPagination',
    'PAGE_SIZE': 10,
    'DEFAULT_FILTER_BACKENDS': ('django_filters.rest_framework.DjangoFilterBackend','rest_framework.filters.SearchFilter','rest_framework.filters.OrderingFilter'),
    'SEARCH_PARAM': 'q'
}
from datetime import timedelta
SIMPLE_JWT = {
    'ACCESS_TOKEN_LIFETIME': timedelta(days=30),
    'REFRESH_TOKEN_LIFETIME': timedelta(days=365),
}

CORS_ALLOW_CREDENTIALS = True
CORS_ORIGIN_ALLOW_ALL = True
CORS_ORIGIN_WHITELIST = ['*']

CORS_ALLOW_METHODS = (
    'DELETE',
    'GET',
    'OPTIONS',
    'PATCH',
    'POST',
    'PUT',
    'VIEW',
)

CORS_ALLOW_HEADERS = (
    'accept',
    'accept-encoding',
    'authorization',
    'content-type',
    'dnt',
    'origin',
    'user-agent',
    'x-csrftoken',
    'x-requested-with',
)


# 科技网登录设置
DOMAIN = "www.abc.cn"
ESCIENCE_APP_KEY = 'testestesttesteste'
ESCIENCE_APP_SECRET = 'testestesttesteste'
ESCIENCE_CALLBACK = 'http://%s/callback/escience' % DOMAIN
ESCIENCE_LOGIN_URL = 'https://passport.escience.cn/oauth2/authorize?response_type=code&redirect_uri=%s&client_id=%s' % (
                ESCIENCE_CALLBACK, ESCIENCE_APP_KEY)
ESCIENCE_LOGIN_HOST = "passport.escience.cn"
ESCIENCE_TOKEN_URL = "https://passport.escience.cn/oauth2/token"



import traceback
from uuid import UUID
from threading import Thread
from django.conf import settings
from django.db import transaction
from django.db.models import Max
from django.utils.timezone import now
from django.dispatch import receiver
from django.db.models.signals import pre_save, post_save, pre_delete, post_delete
from .models import Cloud, Image, Instance, Volume, Mount, InstanceOperation, Group, GroupOperation
from .models import INSTANCE_STATUS, INSTANCE_OPERATION, OPERATION_STATUS, VOLUME_STATUS
from . import utils
from user.utils import get_space

#TODO use singleton

from django.dispatch import Signal
materialized = Signal(providing_args=["instance","name"])
tidied = Signal(providing_args=["instance","name"])
selected = Signal(providing_args=["instance","name"])
from .models import bootstraped, monitored, executed, destroyed

@receiver(materialized)
@receiver(destroyed)
@receiver(monitored)
@receiver(tidied)
@receiver(selected)
@receiver(executed)
@receiver(bootstraped)
def log(sender,instance,name,**kwargs):
    print('SIGNAL INFO:{}/{}/{}'.format(sender._meta.app_label, sender._meta.verbose_name, instance), name)

@receiver(post_save, sender=Cloud)
def bootstrap_cloud(sender,instance,**kwargs):
    if kwargs['created']:
        instance.import_image()
        instance.import_template()
        (pubkey,prikey)=utils.gen_ssh_key()
        pubkey+=' '+instance._key_name
        instance.instance_credential_private_key=prikey
        instance.driver.keypairs.create(name=instance._key_name, public_key=pubkey)
        instance.save()
        instance.bootstrap()

@receiver(pre_delete, sender=Cloud)
def cleanup_cloud(sender,instance,**kwargs):
    instance.driver.keypairs.delete(instance._key_name)

@receiver(post_save, sender=Image)
def clone_image(sender,instance,**kwargs):
    if not instance.parent: return
    if instance.parent.access_id != instance.access_id:
        raise Exception('Must keep access_id same with parent')
    if instance.parent.cloud != instance.cloud:
        raise Exception('Must keep cloud same with parent')
    if instance.min_ram<instance.parent.min_ram:
        sender.objects.filter(pk=instance.pk).update(min_ram=instance.parent.min_ram)
    if instance.min_disk<instance.parent.min_disk:
        sender.objects.filter(pk=instance.pk).update(min_disk=instance.parent.min_disk)

@receiver(pre_delete, sender=Image)
def destroy_image(sender,instance,**kwargs):
    if not instance.protected:
        if sender.objects.filter(access_id=instance.access_id).count()==1:# avoid to delete shared cloud images
            instance.cloud.driver.images.delete(instance.access_id)

# actions relies on status must be registered to the monitored signal first.
@receiver(materialized, sender=Instance)
@receiver(post_save, sender=Mount)
@receiver(tidied, sender=InstanceOperation)
@receiver(executed, sender=InstanceOperation)
def monitor_instance(sender, instance, **kwargs):
    if sender==Mount:
        if not kwargs['created'] or instance.ready: return
        instance=instance.instance
    if sender==InstanceOperation:
        if instance.status==OPERATION_STATUS.waiting.value: return
        instance=instance.target
    if not instance.ready: return
    Thread(target=instance.monitor).start()

@receiver(post_save, sender=Instance)
def materialize_instance(sender, instance, **kwargs):
    if not kwargs['created'] or instance.ready: return
    instance.built_time=now()
    instance.save()
    instance.update_remedy_script(instance.template.remedy_script+'\n'+instance.image.remedy_script)
    def materialize(instance=instance):
        remark = settings.PACKONE_LABEL+'.'+instance.cloud.name+';'
        if instance.remark: remark+=instance.remark
        ins=instance.cloud.driver.instances.create(
            instance.image.access_id,
            instance.template.access_id,
            remark
        )
        instance=sender.objects.get(pk=instance.pk)
        instance.uuid=UUID(ins.id.replace('-', ''), version=4)
        instance.built_time=ins.created
        instance.ipv4=ins.addresses['provider'][0]['addr']
        instance.save()
        hosts='###instance###\n'+instance.hosts_record
        if instance.cloud.hosts: hosts=hosts+'\n###cloud###\n'+instance.cloud.hosts
        instance.update_remedy_script(utils.remedy_script_hosts_add(hosts, overwrite=True),heading=True)
        instance.set_password()
        instance.set_public_key()
        instance.remedy(manual=False)
        materialized.send(sender=sender, instance=instance, name='materialized')
    Thread(target=materialize).start()

@receiver(pre_save, sender=Instance)
def update_instance_hostname(sender, instance, **kwargs):
    if not instance.pk:
        if not instance.hostname:
            instance.hostname=instance.image.hostname
        instance.remedy_script_todo+='\n'+utils.remedy_script_hostname(instance.hostname)
        return
    old=sender.objects.get(pk=instance.id)
    if old.hostname!=instance.hostname:
        instance.remedy(
            script=utils.remedy_script_hostname(instance.hostname),
            manual=False
        )

@receiver(pre_delete, sender=Instance)
def destroy_instance(sender,instance,**kwargs):
    #to aviold repeated deletion
    for instance in sender.objects.select_for_update().filter(pk=instance.pk):
        def destroy():
            if not instance.ready:
                print('WARNNING: delete instance under building')
            else:
                try:
                    instance.cloud.driver.instances.force_delete(str(instance.uuid))
                except Exception as e:#TODO may spam the log
                    instance.pk=None
                    instance.save()
                    traceback.print_exc()
                    return
            destroyed.send(sender=sender, instance=instance, name='destroyed')
        transaction.on_commit(Thread(target=destroy).start)

@receiver(post_save, sender=Volume)
def materialize_volume(sender, instance, **kwargs):
    if not kwargs['created'] or instance.ready: return
    instance.built_time=now()
    instance.save()
    @transaction.atomic
    def materialize(volume=instance):
        volume=sender.objects.select_for_update().get(pk=volume.pk)
        remark = settings.PACKONE_LABEL+'.'+volume.cloud.name+';'
        if volume.remark: remark+=volume.remark
        info=volume.cloud.driver.volumes.create(
            volume.capacity,
            remark=remark
        )
        volume.uuid=UUID(info.id.replace('-', ''), version=4)
        volume.built_time=now()
        volume.status=VOLUME_STATUS.available.value
        volume.save()
        materialized.send(sender=sender, instance=volume, name='materialized')
    transaction.on_commit(Thread(target=materialize).start)

@receiver(pre_delete, sender=Volume)
def destroy_volume(sender,instance,**kwargs):
    #to aviold repeated deletion
    for volume in sender.objects.select_for_update().filter(pk=instance.pk):
        def destroy():
            if not volume.ready:
                print('WARNNING: delete volume under building')
            else:
                try:
                    volume.cloud.driver.volumes.delete(
                        str(volume.uuid)
                    )
                except Exception as e:#TODO may spam the log
                    volume.pk=None
                    volume.save()
                    traceback.print_exc()
                    return
            destroyed.send(sender=sender, instance=volume, name='destroyed')
        transaction.on_commit(Thread(target=destroy).start)

@receiver(monitored, sender=Instance)
@receiver(materialized, sender=Volume)
@transaction.atomic
def mount(sender, instance, **kwargs):
    if instance.deleting: return
    instance=sender.objects.select_for_update().get(pk=instance.pk)
    if not instance.mountable: return
    mounts=instance.mount_set.select_for_update().filter(
        completed_time=None,
        volume__status = VOLUME_STATUS.available.value
    ) if sender==Instance else Mount.objects.select_for_update().filter(completed_time=None, volume=instance)
    if not mounts.exists(): return
    @transaction.atomic
    def materialize(mount):
        mount=Mount.objects.select_related('volume').select_for_update().get(pk=mount.pk)
        vol=mount.volume.cloud.driver.volumes.mount(
            str(mount.volume.uuid),
            str(mount.instance.uuid)
        )
        mount.dev=vol.attachments[0]['device']#TODO allow multiple mounts for the same volume
        mount.completed_time=now()
        mount.save()
        mount.instance.update_remedy_script(
            utils.remedy_script_mount_add(mount),
            heading=True
        )
        mount.volume.status=VOLUME_STATUS.mounted.value
        mount.volume.save()
        materialized.send(sender=Mount, instance=mount, name='materialized')
    for mount in mounts:
        if not mount.instance.mountable: continue
        mount.completed_time=now()
        mount.save()
        Thread(target=materialize,args=(mount,)).start()

@receiver(pre_delete, sender=Mount)
def umount(sender,instance,**kwargs):
    #to aviold repeated deletion
    for mount in sender.objects.select_for_update().filter(pk=instance.pk):
        @transaction.atomic
        def destroy(mount=mount):
            volume=Volume.objects.select_for_update().get(pk=mount.volume.pk)
            if not mount.ready:
                print('WARNNING: delete mount under building')
            else:
                try:
                    mount.volume.cloud.driver.volumes.unmount(
                        str(mount.volume.uuid),
                        str(mount.instance.uuid)
                    )
                except Exception as e:
                    mount.pk=None
                    mount.save()
                    traceback.print_exc()
                    return
                volume.status=VOLUME_STATUS.available.value
                volume.save()
                mount.instance.update_remedy_script(utils.remedy_script_mount_remove(mount))
            destroyed.send(sender=sender, instance=mount, name='destroyed')
        transaction.on_commit(Thread(target=destroy).start)

@receiver(materialized, sender=Instance)#TODO instance may created before be added to group
@receiver(materialized, sender=Mount)
@transaction.atomic
def materialize_group(sender,instance,**kwargs):
    if sender==Mount: instance=instance.instance
    elif instance.mount_set.all().exists(): return
    for group in instance.group_set.select_for_update():
        if group.ready: continue
        if sender==Mount and group.mounts.filter(dev=None).exists(): continue
        if sender==Instance and group.instances.filter(uuid=None).exists(): continue
        group.hosts = '###group {}###\n'.format(group.long_id)+'\n'.join([ins.hosts_record for ins in group.instances.all()])
        group.built_time=now()
        group.save()
        for ins in group.instances.all():
            ins.remedy(manual=False)
        GroupOperation(
            operation=INSTANCE_OPERATION.start.value,
            target=group,
            status=OPERATION_STATUS.running.value,
            ignore_error=True
        ).save()
        group.remedy(utils.remedy_script_hosts_add(group.hosts),manual=False)
        materialized.send(sender=Group, instance=group, name='materialized')

@receiver(destroyed, sender=Instance)
@transaction.atomic
def destroy_group(sender,instance,**kwargs):
    for group in Group.objects.select_for_update().filter(
        deleting=True,
    ):
        if not group.instances.all().exists():
            destroyed.send(sender=Group, instance=group, name='destroyed')
            group.delete()

@receiver(monitored, sender=Instance)
@receiver(tidied, sender=GroupOperation)
@receiver(executed, sender=GroupOperation)
def monitor_group(sender, instance, **kwargs):
    if sender==Instance:
        if instance.deleting: return
        for group in instance.group_set.all():
            status=group.instances.all().aggregate(Max('status'))['status__max']#TODO use join
            Group.objects.filter(pk=group.pk).update(status=status)
            group.refresh_from_db()
            monitored.send(sender=Group, instance=group, name='monitored')
    else:
        if instance.status==OPERATION_STATUS.waiting.value: return
        instance.target.monitor()

@receiver(post_save, sender=InstanceOperation)
@receiver(post_save, sender=GroupOperation)
def tidy_operation(sender,instance,created,**kwargs):
    if not created or instance.serial: return #only following serial operations will not be tidied
    if instance.operation==INSTANCE_OPERATION.remedy.value and instance.script and not instance.tidied:
        supervisor_ops=[s.value for s in INSTANCE_OPERATION]
        ops=utils.remedy_script_tidy(instance.script,supervisor_ops)
        for i in range(len(ops)):
            op=ops[i]
            if i==0:
                if op in supervisor_ops:
                    instance.script=None
                    instance.operation=op
                    instance.tidied=True
                else:
                    instance.script=op
                    instance.tidied=True
                instance.save()
            else:
                if op in supervisor_ops:
                    op_instance=sender(
                        target=instance.target,
                        operation=op,
                    )
                else:
                    op_instance=sender(
                        target=instance.target,
                        operation=INSTANCE_OPERATION.remedy.value,
                        script=op,
                    )
                op_instance.status=OPERATION_STATUS.waiting.value
                op_instance.serial=instance
                op_instance.tidied=True
                op_instance.manual=False
                op_instance.save()
    tidied.send(sender=sender, instance=instance, name='tidied')

@receiver(monitored, sender=Instance)
@receiver(monitored, sender=Group)
@transaction.atomic#to aviod deleted target and duplicated remedy
def select_operation(sender,instance,**kwargs):
    for target in instance.__class__.objects.select_for_update().filter(pk=instance.pk):
        target.remedy(manual=False)
        ops=target.get_next_operations().select_for_update()
        if not ops.exists(): return
        for op in ops:
            if op.runnable:
                op.status=OPERATION_STATUS.running.value
                op.started_time=now()
                op.completed_time=None
                op.save()
                selected.send(sender=op.__class__, instance=op, name='selected')
                break
            else:
                op.status=OPERATION_STATUS.waiting.value
                op.save()

@receiver(selected, sender=InstanceOperation)
@receiver(selected, sender=GroupOperation)
def execute_operation(sender,instance,**kwargs):
    instance.execute()

@receiver(executed, sender=InstanceOperation)
@transaction.atomic
def close_group_operation(sender, instance, **kwargs):
    for running_op in GroupOperation.objects.select_for_update().filter(
        batch_uuid=instance.batch_uuid,
        started_time__isnull=False
    ):
        if not running_op.get_remain_oprations().exists():
            running_op.completed_time=now()
            running_op.status=running_op.get_status()
            running_op.save()
            executed.send(sender=GroupOperation, instance=running_op, name='executed')

@receiver(post_delete, sender=InstanceOperation)
def purge_group_operation(sender, instance, **kwargs):
    g_op=GroupOperation.objects.filter(batch_uuid=instance.batch_uuid).first()
    if g_op and not g_op.get_sub_operations().exists():
        g_op.delete()

@receiver(monitored, sender=Instance)
@receiver(destroyed, sender=Mount)
def cleanup(sender,instance,**kwargs):
    if sender==Instance:
        if not instance.deleting: return
        ms=instance.mount_set.all()
        if ms.exists():
            if instance.umountable:
                for m in ms:
                    m.delete()
                instance.remark+=';umounting'
                instance.save()
        else:
            instance.refresh_from_db()
            if not instance.remark.endswith('umounting'):
                instance.delete()
    else:
        if instance.instance.deleting and not instance.instance.mount_set.all().exists():
            instance.instance.delete()
        if instance.volume.deleting:
            instance.volume.delete()


from .h3cloudos import Driver as H3CDriver
from .h3cloudos import ImageManager as H3CImageManager
from .h3cloudos import Image
from .h3cloudos import InstanceManager#TODO  for InstanceManager.mountable_status

class Driver(H3CDriver):
    def __init__(self, cloud):
        if not 'nova-availability_zone' in cloud.platform_credential: cloud.platform_credential['nova-availability_zone']="CASCloud01"
        if not 'cinder-availability_zone' in cloud.platform_credential: cloud.platform_credential['cinder-availability_zone']="cstcloud_cinder01"
        cloud.platform_credential['endpoint']='http://159.226.245.2:9000'
        H3CDriver.__init__(self, cloud)
        self.images=ImageManager(self)

class ImageManager(H3CImageManager):
    def list(self):
        images=[]
        for item in self.driver._get('/v2/images')['images']:
            if item['name']=='CentOS7-YouHua':
                item['name']+='-GenericCloud'
            images.append(Image(item))
        return images


from django.contrib.auth.models import User
from rest_framework import serializers
from . import models

class AccountSerializer(serializers.ModelSerializer):
    last_login = serializers.ReadOnlyField()
    date_joined = serializers.ReadOnlyField()
    class Meta:
        model = User
        fields = ('username', 'email', 'password', 'first_name', 'last_name', 'last_login', 'date_joined')

class ProfileSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Profile
        fields = ('id','organization','remark')

class ProfilePKField(serializers.PrimaryKeyRelatedField):
    def get_queryset(self):
        user = self.context['request'].user
        queryset = models.Profile.objects.filter(account=user)
        return queryset

class BalanceSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Balance
        fields = ('id','cloud','balance','enabled')

class CredentialSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Credential
        fields = ('id','ssh_user','ssh_public_key','ssh_passwd')
